import java.net.ServerSocket;
import java.net.Socket;
import java.util.Vector;
import java.io.IOException;
import java.util.Date;

public class TCPReflector implements ConnectionMonitor
{
   private int                mPortNumber     = -1;
   private String             mDestHost       = null;
   private int                mDestPortNumber = -1;
   private boolean            mEcho           = false;
   private boolean            mQuiet          = false;
   private boolean            mBlock          = false;
   private boolean            mXBlock         = false;
   private ServerSocket       mMainSocket     = null;
   private Vector<Connection> mConnections    = null;
   private int                mID              = 0;

   private static final String border =
      "####################################" +
      "####################################";

   public TCPReflector(String args[]) throws Exception
   {
      boolean error = false;

      try
      {
         processArgs(args);
      }
      catch(Exception e)
      {
         System.out.println("ERROR: " + e.getMessage() + "\n");
         error = true;
      }

      if(mPortNumber <= 0 || mDestPortNumber <= 0 || mDestHost == null)
         error = true;

      if(error)
      {
         usage();
         System.exit(-1);
      }

      mConnections = new Vector<Connection>();
      mMainSocket  = new ServerSocket(mPortNumber);

      printMsg(
         "Starting program: " + getClass().getName(),
         "On port number  : " + mPortNumber,
         "Dest host:port  : " + mDestHost + ":" + mDestPortNumber
      );
   }

   private int parsePort(int i, String args[]) throws Exception
   {
      int value = 0;

      if(i >= args.length)
      {
         throw new Exception(
            "Expected numeric argument after -port"
         );
      }

      try
      {
         value = Integer.parseInt(args[i]);
      }
      catch(NumberFormatException e) {}

      if(value <= 0)
      {
         throw new Exception(
            "Value for -port must be a valid integer > 0 [" + args[i] + "]"
         );
      }

      return(value);
   }

   private void parseDest(int i, String args[]) throws Exception
   {
      if(i >= args.length)
      {
         throw new Exception(
            "Expected argument after -dest"
         );
      }

      int colonIndex = args[i].indexOf(':');

      if(colonIndex == -1)
      {
         throw new Exception(
            "Expected ':' separator in argument to -dest [" + args[i] + "]"
         );
      }

      String s1 = args[i].substring(0, colonIndex).trim();
      String s2 = args[i].substring(colonIndex + 1).trim();

      if(s1.equals(""))
      {
         throw new Exception(
            "Missing host name in argument to -dest [" + args[i] + "]"
         );
      }

      int value = 0;

      try
      {
         value = Integer.parseInt(s2);
      }
      catch(NumberFormatException e) {}

      if(value <= 0)
      {
         throw new Exception(
            "Port number for -dest argument must be an integer > 0 [" +
            args[i] + "]"
         );
      }

      mDestHost       = s1;
      mDestPortNumber = value;
   }

   private void processArgs(String args[]) throws Exception
   {
      int i = 0;

      while(i < args.length)
      {
         String s = args[i].toLowerCase();

         if(s.equals("-echo"))
         {
            mEcho = true;
         }
         else if(s.equals("-quiet"))
         {
            mQuiet = true;
         }
         else if(s.equals("-block"))
         {
            mBlock = true;
         }
         else if(s.equals("-xblock"))
         {
            mXBlock = true;
         }
         else if(s.equals("-port"))
         {
            mPortNumber = parsePort(++i, args);
         }
         else if(s.equals("-dest"))
         {
            parseDest(++i, args);
         }
         else
         {
            throw new Exception(
               "Unrecognized argument: [" + args[i] + "]"
            );
         }

         i++;
      }
   }

   public static void usage()
   {
      System.out.println("Usage: program <args>");
      System.out.println("   -port listenPort");
      System.out.println("   -dest destHost:destPortNumber");
      System.out.println("   -echo");
      System.out.println("   -quiet");
      System.out.println("   -block");
      System.out.println("   -xblock");
      System.out.println("");
      System.out.println("-port and -dest are required");
   }

   private void printMsg(String s1)
   {
      printMsg(new String[] { s1 });
   }

   private void printMsg(String s1, String s2)
   {
      printMsg(new String[] { s1, s2 });
   }

   private void printMsg(String s1, String s2, String s3)
   {
      printMsg(new String[] { s1, s2, s3 });
   }

   private void printMsg(String msg[])
   {
      if(!mQuiet)
      {
         String timeString = "" + System.currentTimeMillis();
         int length        = timeString.length();

         timeString = timeString.substring(0, length - 3)
                    + "." + timeString.substring(length - 4);

         System.out.println(border);
         System.out.println("# " + timeString);

         for(int i = 0; i < msg.length; i ++)
            System.out.println("# " + msg[i]);

         System.out.println(border + "\n");
      }
   }

   public void serve()
   {
      while(true)
      {
         try
         {
            Socket newSocket = mMainSocket.accept();

            new Connection(
               newSocket,
               (ConnectionMonitor)this, 
               mDestHost,
               mDestPortNumber,
               mEcho,
               mBlock,
               mXBlock,
               mID++
            );
         }
         catch(IOException e) {}
      }
   }

   public void attemptingConnection(Connection c)
   {
      printMsg(
         getClass().getName(),
         "Connection initiated from: " + c.getSrcHost()
      );
   }

   public void addConnection(Connection c)
   {
      printMsg(
         getClass().getName(),
         "Connection established from: " + c.getSrcHost()  + ":" +
            mPortNumber, 
         "To                         : " + c.getDestHost() + ":" + 
            c.getDestPort()
      );

      synchronized(mConnections)
      {
         mConnections.addElement(c);
      }
   }

   public void removeConnection(Connection c)
   {
      printMsg(
         getClass().getName(),
         "Removing connection from: " + c.getSrcHost()  + ":" +
            mPortNumber, 
         "To                      : " + c.getDestHost() + ":" + 
            c.getDestPort()
      );

      synchronized(mConnections)
      {
         mConnections.removeElement(c);
      }
   }

   public void connectionError(Connection c, String errMsg)
   {
      printMsg(
         getClass().getName(),
         "Error involving connection from:" + c.getSrcHost()  + ":" +
            mPortNumber, 
         errMsg
      );
   }

   public static void main(String args[]) throws Exception
   {
      TCPReflector t = new TCPReflector(args);
      t.serve();
   }
}
